#
# Copyright (C) 2012 Marlus Saraiva, Rodrigo Maia
# Copyright (C) 2019 Stefan Eilers
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

require "jasper-rails/version"
require "rails"
require "rjb"
#require "action_controller/metal/responder"

if Mime::Type.lookup_by_extension("pdf").nil?
    Mime::Type.register "application/pdf", :pdf
end

module JasperRails

    class << self
        attr_accessor :config
    end

    module Jasper
        module Rails

            def self.classpath
                classpaths = '.'
                Dir["#{File.dirname(__FILE__)}/java/*.jar"].each do |jar|
                    classpaths << File::PATH_SEPARATOR + File.expand_path(jar)
                end

                Dir["lib/*.jar"].each do |jar|
                    classpaths << File::PATH_SEPARATOR + File.expand_path(jar)
                end
                classpaths
            end

            ##
            ## @brief      Render a PDF file based on jasper file and an optional datasource.
            ##
            ## @param      jasper_file  The Jasper-Report Template (.jrxml)
            ## @param      datasource   The Datasource (JRAbstractTextDataSource)
            ## @param      parameters   The parameters (Ruby-Hash)
            ##
            ## @return     The PDF-Data
            ##
            def self.render_pdf(jasper_file, parameters, datasource = nil )
                # The code below is to workaround declaring constants within methods
                # We would like to delay these till a request is received to workaround Apache Passenger issue
                locale                      = Rjb::import 'java.util.Locale'
                jRException                 = Rjb::import 'net.sf.jasperreports.engine.JRException'
                jasperCompileManager        = Rjb::import 'net.sf.jasperreports.engine.JasperCompileManager'
                jasperExportManager         = Rjb::import 'net.sf.jasperreports.engine.JasperExportManager'
                jasperFillManager           = Rjb::import 'net.sf.jasperreports.engine.JasperFillManager'
                jasperPrint                 = Rjb::import 'net.sf.jasperreports.engine.JasperPrint'
                jREmptyDataSource           = Rjb::import 'net.sf.jasperreports.engine.JREmptyDataSource'
                jRXPathQueryExecuterFactory = silence_warnings{Rjb::import 'net.sf.jasperreports.engine.query.JRXPathQueryExecuterFactory'}
                inputSource                 = Rjb::import 'org.xml.sax.InputSource'
                stringReader                = Rjb::import 'java.io.StringReader'
                hashMap                     = Rjb::import 'java.util.HashMap'
                javaString                  = Rjb::import 'java.lang.String'
                jFreeChart                  = Rjb::import 'org.jfree.chart.JFreeChart'
                marschalException           = Rjb::import 'org.exolab.castor.xml.MarshalException'

                options ||= {}
                parameters ||= {}
                jrxml_file  = jasper_file.sub(/\.jasper$/, ".jrxml")

                begin
                    # Default report params
                    config = {
                        :report_params=>{
                            "REPORT_LOCALE"    => locale.new('de', 'DE'),
                            "XML_LOCALE"       => locale.new('de', 'DE'),
                            "XML_DATE_PATTERN" => 'yyyy-MM-dd'
                        }
                    }

                    # Converting default report params to java HashMap
                    jasper_params = hashMap.new
                    config[:report_params].each do |k,v|
                        jasper_params.put(k, v)
                    end

                    # Convert the ruby parameters' hash to a java HashMap, but keeps it as
                    # default when they already represent a JRB entity.
                    # Pay attention that, for now, all other parameters are converted to string!
                    parameters.each do |key, value|
                        jasper_params.put(javaString.new(key.to_s), parameter_value_of(value))
                    end

                    # Compile it, if needed
                    if !File.exist?(jasper_file) || (File.exist?(jrxml_file) && File.mtime(jrxml_file) > File.mtime(jasper_file))
                        jasperCompileManager.compileReportToFile(jrxml_file, jasper_file)
                    end

                    # Fill the report
                    if datasource
                        jasper_print = jasperFillManager.fillReport(jasper_file, jasper_params, datasource || jREmptyDataSource.new)
                    else
                        jasper_print = jasperFillManager.fillReport(jasper_file, jasper_params)
                    end
                    # Export it!
                    jasperExportManager._invoke('exportReportToPdf', 'Lnet.sf.jasperreports.engine.JasperPrint;', jasper_print)
                rescue Exception=>e
                    if e.respond_to? 'printStackTrace'
                        ::Rails.logger.error e.message
                        e.printStackTrace
                    else
                        ::Rails.logger.error e.message + "\n " + e.backtrace.join("\n ")
                    end
                    raise e
                end
            end

            ##
            ## @brief      Render a Jasper-Report by using a JsonDataSource
            ##
            ## @param      jasper_file  The jasper file (.jrxml)
            ## @param      parameters   The parameters (Ruby-Hash)
            ## @param      json_data    The json data (JSON-String)
            ##
            ## @return     { description_of_the_return_value }
            ##
            def self.render_json_to_pdf(jasper_file, parameters, json_data)
                setup_jasper

                # Convert json string to byte array.
                jsonDataStream = byteArrayInputStream.new(javaString.new(json_data).getBytes());
                data_source = jsonDataSource.new(jsonDataStream)

                render_pdf(jasper_file, parameters, data_source)
            end

            ##
            ## @brief      Render a report that contains a datasource.
            ##
            ## @param      jasper_file  The jasper file
            ## @param      parameters   The parameters
            ##
            ## @return     The rendered pdf file.
            ##
            def self.render_datasource_to_pdf(jasper_file, parameters)
                setup_jasper

                render_pdf(jasper_file, parameters, nil)
            end

            private

            def self.setup_jasper
                ENV['CLASS_PATH'] ||= classpath
                ENV['JVM_ARGS'] ||= '-Djava.awt.headless=true,-Xms128M,-Xmx256M'

                Rjb::load( ENV['CLASS_PATH'], ENV['JVM_ARGS'].split(",") ) unless Rjb::loaded?

                # The code below is to workaround declaring constants within methods
                # We would like to delay these till a request is received to workaround Apache Passenger issue
                jREmptyDataSource           = Rjb::import 'net.sf.jasperreports.engine.JREmptyDataSource'
                jsonDataSource              = Rjb::import 'net.sf.jasperreports.engine.data.JsonDataSource'
                byteArrayInputStream        = Rjb::import 'java.io.ByteArrayInputStream'
                jsonNode 					= Rjb::import 'com.fasterxml.jackson.databind.JsonNode'
                javaString                  = Rjb::import 'java.lang.String'


            end

            # Returns the value without conversion when it's converted to Java Types.
            # When isn't a Rjb class, returns a Java String of it.
            def self.parameter_value_of(param)
                javaString                  = Rjb::import 'java.lang.String'
                # Using Rjb::import('java.util.HashMap').new, it returns an instance of
                # Rjb::Rjb_JavaProxy, so the Rjb_JavaProxy parent is the Rjb module itself.
                if param.class.parent == Rjb
                    param
                else
                    javaString.new(param.to_s)
                end
            end
        end
    end
end
